from importlib.metadata import version
import traceback

from flask import Flask, request
from werkzeug.exceptions import HTTPException, NotFound, UnsupportedMediaType

from .invoke import invoke


def create_app():
    app = Flask(__name__)

    @app.route("/")
    def route_index():
        return {
            "app": "Rippling Flux Server",
            "versions": {dep: version(dep) for dep in ["Flask", "rippling-flux-sdk", "rippling-flux-dev-tools"]}
        }

    @app.route("/invoke", methods=["POST"])
    def route_invoke():
        return invoke(request.get_json())

    @app.errorhandler(404)
    def page_not_found(err: NotFound):
        return err.description, err.code

    @app.errorhandler(415)
    def page_unsupported_media_type(err: UnsupportedMediaType):
        return err.description, err.code

    @app.errorhandler(HTTPException)
    def errorhandler_catch(err: HTTPException):
        return traceback.format_exc(), err.code

    return app


import io
import json
import unittest
from decimal import Decimal
from enum import Enum

from flux_dev_tools.server.serialization import FluxJSONDecoder, FluxJSONEncoder


class Gender(Enum):
    MALE = "Male"
    FEMALE = "Female"


class Address:
    street: str
    city: str
    state: str
    zip_code: str


class Employee:
    name: str
    age: int
    salary: Decimal
    gender: Gender
    address: Address


class TestCustomJSONEncoderDecoder(unittest.TestCase):
    def assertEmployeeEqual(self, employee, decoded_employee):
        self.assertEqual(employee.name, decoded_employee.name)
        self.assertEqual(employee.age, decoded_employee.age)
        self.assertEqual(employee.salary, decoded_employee.salary)
        self.assertEqual(employee.gender, decoded_employee.gender)
        self.assertEqual(employee.address.street, decoded_employee.address.street)
        self.assertEqual(employee.address.city, decoded_employee.address.city)
        self.assertEqual(employee.address.state, decoded_employee.address.state)
        self.assertEqual(employee.address.zip_code, decoded_employee.address.zip_code)

    def test_simple_types_encoding_and_decoding(self):
        # Test encoding and decoding of simple types
        text = "Hello World"
        encoded_text = json.dumps(text, cls=FluxJSONEncoder)
        decoded_text = json.loads(encoded_text, cls=FluxJSONDecoder, target_type=str)
        self.assertEqual(decoded_text, text)

        number = 123
        encoded_number = json.dumps(number, cls=FluxJSONEncoder)
        decoded_number = json.loads(encoded_number, cls=FluxJSONDecoder, target_type=int)
        self.assertEqual(decoded_number, number)

        decimal = Decimal("123.45")
        encoded_decimal = json.dumps(decimal, cls=FluxJSONEncoder)
        decoded_decimal = json.loads(encoded_decimal, cls=FluxJSONDecoder, target_type=Decimal)
        self.assertEqual(decoded_decimal, decimal)

    def test_custom_class_encoding_and_decoding(self):
        # Test encoding and decoding of custom class instances
        employee = Employee()
        employee.name = "Alice"
        employee.age = 25
        employee.salary = Decimal("60000.75")
        employee.gender = Gender.FEMALE
        address = Address()
        address.street = "123 Main St"
        address.city = "New York"
        address.state = "NY"
        address.zip_code = "10001"
        employee.address = address

        encoded_employee = json.dumps(employee, cls=FluxJSONEncoder)
        decoded_employee = json.loads(encoded_employee, cls=FluxJSONDecoder, target_type=Employee)

        self.assertEmployeeEqual(decoded_employee, employee)

    def test_enum_encoding_and_decoding(self):
        # Test encoding and decoding of Enums
        encoded_data = json.dumps(Gender.FEMALE, cls=FluxJSONEncoder)
        decoded_data = json.loads(encoded_data, cls=FluxJSONDecoder, target_type=Gender)

        self.assertEqual(decoded_data, Gender.FEMALE)

    def test_string_io_encoding_and_decoding(self):
        # Test encoding and decoding of StringIO
        sample_data = "Hello World"
        stream = io.StringIO(sample_data)

        encoded_string_io = json.dumps(stream, cls=FluxJSONEncoder)
        decoded_string_io = json.loads(encoded_string_io, cls=FluxJSONDecoder, target_type=io.IOBase)

        decoded_data = decoded_string_io.read()
        self.assertEqual(decoded_data, sample_data)

    def test_bytes_io_encoding_and_decoding(self):
        # Test encoding and decoding of BytesIO
        sample_data = b"\x01\x02\x03\x04"
        stream = io.BytesIO(sample_data)

        encoded_bytes_io = json.dumps(stream, cls=FluxJSONEncoder)
        decoded_bytes_io = json.loads(encoded_bytes_io, cls=FluxJSONDecoder, target_type=io.IOBase)

        decoded_data = decoded_bytes_io.read()
        self.assertEqual(decoded_data, sample_data)

    def test_list_of_objects_encoding_and_decoding(self):
        # Test encoding and decoding of lists of objects
        employees = [Employee(), Employee(), Employee()]
        employees[0].name = "Employee 1"
        employees[1].name = "Employee 2"
        employees[2].name = "Employee 3"

        encoded_employees = json.dumps(employees, cls=FluxJSONEncoder)
        decoded_employees = json.loads(encoded_employees, cls=FluxJSONDecoder, target_type=list[Employee])

        for decoded_employee, original_employee in zip(decoded_employees, employees):
            self.assertEqual(decoded_employee.name, original_employee.name)

    def test_list_of_simple_types_encoding_and_decoding(self):
        # Test encoding and decoding of lists of simple types
        numbers = [1, 2, 3, 4, 5]
        encoded_numbers = json.dumps(numbers, cls=FluxJSONEncoder)
        decoded_numbers = json.loads(encoded_numbers, cls=FluxJSONDecoder, target_type=list[int])

        self.assertEqual(decoded_numbers, numbers)

    def test_dict_of_simple_types_encoding_and_decoding(self):
        # Test encoding and decoding of dicts of simple types
        data = {"a": 1, "b": 2, "c": 3}
        encoded_data = json.dumps(data, cls=FluxJSONEncoder)
        decoded_data = json.loads(encoded_data, cls=FluxJSONDecoder, target_type=dict[str, int])

        self.assertEqual(decoded_data, data)


import importlib
import json
from functools import reduce

from .serialization import FluxJSONEncoder, FluxJSONDecoder


def invoke(event):
    hook = event['hook']
    hook_params = event['hook_params']
    app_implementation_type = event['app_implementation_type']
    app_name = event['app_name']
    kit_name = event['kit_name']
    capability_snake_case = convert_to_snakecase(app_implementation_type)
    kit_snake_case = convert_to_snakecase(kit_name)

    module_path = ".".join(
        [
            "app",
            kit_snake_case,
            "capabilities",
            capability_snake_case,
            "implementation",
        ]
    )

    module = importlib.import_module(module_path)

    module_cls = app_implementation_type + "Impl"

    if module and hasattr(module, module_cls):
        app_class = getattr(module, module_cls)
        hook_method = getattr(app_class, hook)

        import inspect
        method_signature = inspect.signature(hook_method)

        parameters = ()
        for parameter_name, parameter in method_signature.parameters.items():
            parameters += (json.loads(hook_params[parameter_name], cls=FluxJSONDecoder, target_type=parameter.annotation),)

    return json.dumps(hook_method(*parameters), cls=FluxJSONEncoder)


def convert_to_snakecase(s: str) -> str:
    """
    Convert a string to snake case. For example, "MyApp" becomes "my_app".
    This is to follow the python convention for module names.
    """
    return reduce(lambda x, y: x + ("_" if y.isupper() else "") + y, s).lower()


import base64
import io
import json
from dataclasses import is_dataclass
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Any, Union, get_args, get_origin


class FluxJSONEncoder(json.JSONEncoder):
    """
    Custom JSON encoder designed for the Flux project.

    This encoder extends the functionality of the default JSONEncoder to handle various data types commonly used
    within the Flux project, ensuring consistent and reliable JSON serialization.

    Supported data types:
    - Simple types: str, datetime, date, Decimal, Enum, bytes, io.IOBase
    - Custom classes with fields of the above simple types
    - Lists of objects, including nested lists
    - Dictionaries with values of the above supported types
    - Optional types, ensuring that optional values are correctly encoded

    Note: This encoder is designed to maintain encoding consistency across the Rippling main application and the Lambda
    function hosting third-party code.

    Usage Example:
    ```
    import json
    from apps.flux_runtime.app_implementation.serialization import FluxJSONEncoder

    encoded_data = json.dumps(data, cls=FluxJSONEncoder)
    ```

    """

    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, Decimal):
            return str(obj)
        if isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, bytes):
            return self._encode_bytes(obj)
        if isinstance(obj, io.IOBase):
            return self._encode_io_base(obj)
        if isinstance(obj, list):
            return [self.default(item) for item in obj]
        if isinstance(obj, dict):
            return {key: self.default(value) for key, value in obj.items()}
        if hasattr(obj, "__dict__"):
            serialized_dict = {"target_type": repr(type(obj))}
            for key, value in obj.__dict__.items():
                serialized_dict[key] = json.dumps(value, cls=FluxJSONEncoder)
            return serialized_dict
        return super().default(obj)

    def _encode_bytes(self, obj):
        if hasattr(obj, "read") and callable(obj.read):
            # If it's an IOBase object, read it and encode as bytes
            return base64.b85encode(obj.read()).decode("utf-8")
        else:
            # Encode bytes directly
            return base64.b85encode(obj).decode("utf-8")

    def _encode_io_base(self, obj):
        if isinstance(obj, io.BytesIO):
            content = obj.getvalue()
            encoding = "bytes"
        elif isinstance(obj, io.StringIO):
            content = obj.getvalue().encode("utf-8")
            encoding = "utf-8"
        else:
            raise TypeError("Unsupported IOBase type")

        encoded_content = base64.b85encode(content)
        return {
            "encoding": encoding,
            "content": encoded_content.decode("utf-8"),
        }


class FluxJSONDecoder(json.JSONDecoder):
    """
    Custom JSON decoder designed for the Flux project.

    This decoder extends the functionality of the default JSONDecoder to handle decoding of JSON strings generated
    using the FluxJSONEncoder, ensuring consistent and reliable JSON deserialization.

    Supported data types:
    - Simple types: str, datetime, date, Decimal, Enum, bytes, io.IOBase
    - Custom classes with fields of the above simple types
    - Lists of objects, including nested lists
    - Dictionaries with values of the above supported types
    - Optional types, ensuring that optional values are correctly decoded

    Note: This decoder is designed to maintain decoding consistency across the Rippling main application and the Lambda
    function hosting third-party code.

    Usage Example:
    ```
    import json
    from apps.flux_runtime.app_implementation.serialization import FluxJSONDecoder

    encoded_data = "..."  # Your encoded JSON data
    decoded_data = json.loads(encoded_data, cls=FluxJSONDecoder, target_type=YourClass)
    ```

    """

    def __init__(self, target_type=None, **kwargs):
        self.target_type = target_type
        super().__init__(**kwargs)

    def decode(self, s, *args, **kwargs):
        obj = super().decode(s, *args, **kwargs)
        return self._convert_object(obj, self.target_type)

    def _convert_object(self, obj, target_type):
        passed_in_target_type = obj.pop("target_type", None) if isinstance(obj, dict) else None
        if get_origin(target_type) is Union:
            for union_type in get_args(target_type):
                if repr(union_type) == passed_in_target_type:
                    target_type = union_type
                    break
        if get_origin(target_type) is Union:
            target_type = type(obj)
        if get_origin(target_type) is list:
            target_type = get_args(target_type)[0]
        if get_origin(target_type) is tuple and obj is not None:
            return tuple(self._convert_object(item, get_args(target_type)[i]) for i, item in enumerate(obj))
        if isinstance(obj, list):
            return [self._convert_object(item, target_type) for item in obj]

        if target_type is not None:
            if obj is None or type(obj) is target_type:
                return obj
            elif target_type is Decimal:
                return Decimal(obj)
            elif target_type is datetime:
                return datetime.fromisoformat(obj)
            elif target_type is date:
                return date.fromisoformat(obj)
            elif target_type is bytes:
                return base64.b85decode(obj)
            elif target_type is io.IOBase:
                return self._decode_io_base(obj)
            elif issubclass(target_type, Enum):
                return target_type(obj)
            elif target_type is float:
                return float(obj)
            elif target_type is int:
                return int(obj)
            elif target_type is str:
                return str(obj)
            elif target_type is dict or get_origin(target_type) is dict:
                deserialized_dict = {}
                for key, value in obj.items():
                    value_target_type = None
                    if (
                        hasattr(target_type, "__args__")
                        and len(target_type.__args__) == 2
                        and target_type.__args__[1] is not Any
                    ):
                        value_target_type = target_type.__args__[1]
                    deserialized_dict[key] = self._convert_object(value, target_type=value_target_type)
                return deserialized_dict
            else:
                if not obj:
                    return target_type()
                deserialized_dict = {}
                for key, value in obj.items():
                    if key in target_type.__annotations__:
                        expected_type = target_type.__annotations__[key]
                        if self.is_optional_type(expected_type):
                            expected_type = expected_type.__args__[0]
                        deserialized_dict[key] = json.loads(value, cls=FluxJSONDecoder, target_type=expected_type)
                if is_dataclass(target_type):
                    return target_type(**deserialized_dict)
                instance = target_type()
                for key, value in target_type.__annotations__.items():
                    setattr(instance, key, deserialized_dict.get(key, None))
                return instance

        if isinstance(obj, str):
            try:
                return datetime.fromisoformat(obj)
            except ValueError:
                pass

        if isinstance(obj, dict):
            return {key: self._convert_object(value, target_type) for key, value in obj.items()}

        return obj

    def is_optional_type(self, typ):
        return get_origin(typ) is Union and type(None) in get_args(typ)

    def _decode_io_base(self, obj):
        encoding = obj["encoding"]
        content = base64.b85decode(obj["content"])

        try:
            if encoding == "bytes":
                return io.BytesIO(content)
            elif encoding == "utf-8":
                return io.StringIO(content.decode("utf-8"))
            else:
                raise TypeError(f"Unsupported encoding type {encoding}")
        except (ImportError, AttributeError):
            raise ValueError("Invalid IOBase type during decoding")